(* Unpack Zip file without using the directory
   (e.g. for damaged or partial zip files)

     Dr. J. Rathlev, D-24222 Schwentinental (kontakt(a)rathlev-home.de)

   Acknowledgements:
     ZLibEx and ZLib routines from http://www.base2ti.com
     AES functions from http://fp.gladman.plus.com/index.htm

   The contents of this file may be used under the terms of the
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - April. 2006
   Vers. 2 - Juni 2007 - handles encrypted zip files
         2.2 - Oct. 2010 - ported to Delphi 2009
   last modified: March 2018  (Delphi 10)
   *)

unit UnZipmain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ComCtrls,
  Vcl.Buttons, FileConsts, ExtFileTools, ZLibEx, AesLib;

const
  ProgName = 'Unpack Zip Files';
  Vers = '2.6';
  CopRgt = ' 2012-2020 Dr. J. Rathlev, 24222 Schwentinental';
  EmailAdr = 'kontakt(a)rathlev-home.de';

  TempName = 'unzip.tmp';
  MaxMemSize = 10*1024*1024;  // max. size of temp. memory stream

type
  TPkFileRef = class(TObject)
    Offset  : int64;
    constructor Create (Ofs : int64);
    end;

  THauptForm = class(TForm)
    lvFiles: TListView;
    Label1: TLabel;
    edZip: TEdit;
    sbLoadZip: TSpeedButton;
    OpenDialog: TOpenDialog;
    bnUnzip: TButton;
    bnEnde: TButton;
    pbUnzip: TProgressBar;
    lbUnzip: TLabel;
    bnCancel: TButton;
    bnInfo: TBitBtn;
    cbCheck: TCheckBox;
    bnUpdate: TButton;
    bnSelectAll: TButton;
    procedure sbLoadZipClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure bnUnzipClick(Sender: TObject);
    procedure bnEndeClick(Sender: TObject);
    procedure bnCancelClick(Sender: TObject);
    procedure bnInfoClick(Sender: TObject);
    procedure bnSelectAllClick(Sender: TObject);
    procedure bnUpdateClick(Sender: TObject);
  private
    { Private-Deklarationen }
    ProgVersName,
    ProgVersDate,
    UserPath,AppPath : string;
    FBufSize         : cardinal;
    FBuffer          : array of byte;
    Total,Prog       : int64;
    fc               : cardinal;
    Base             : string;
    Cancelled        : boolean;
    Password         : AnsiString;
    FEncryption      : TEncryption;
    function GzUnCompress (fSource,fDest : TStream; sLength : int64; var FCrc : cardinal) : integer;
    function ReadZipFileHeader (sZip : TFileStream; var ZipInfo : TZipFileInfo) : boolean;
  public
    { Public-Deklarationen }
    procedure ReadZip (FName : string);
  end;

var
  HauptForm: THauptForm;

implementation

{$R *.dfm}

uses System.StrUtils, WinShell, WinApiUtils, Winapi.ShlObj, WinUtils, ShellDirDlg,
  Web.HttpApp, PwdDlg, GnuGetText, XlFileUtils, LangUtils, FileUtils, InitProg,
  FileErrors;

constructor TPkFileRef.Create (Ofs : int64);
begin
  inherited Create;
  Offset:=Ofs;
  end;

function ZipCrc32(Crc : cardinal; const buf; len: Integer): cardinal;
begin
  Result:=Crc32 (Crc,pbyte(@buf),len);
  end;

procedure THauptForm.FormCreate(Sender: TObject);
begin
  TranslateComponent(self);
  InitPaths(AppPath,UserPath);
  InitVersion(_('Unpacking of damaged Zip archives'),Vers,CopRgt,3,3,ProgVersName,ProgVersDate);
  Caption:=_('Unpacking of damaged Zip archives')+' ('+VersInfo.Comments+')';
  SetLength(FBuffer,defBufferSize);
  FBufSize:=defBufferSize;
  Base:=UserPath;
  Total:=0; Prog:=0; fc:=0;
  bnCancel.Enabled:=false;
  FEncryption:=nil;
  end;

procedure THauptForm.FormDestroy(Sender: TObject);
begin
  FBuffer:=nil;
  end;

function THauptForm.GzUnCompress (fSource,fDest : TStream; sLength : int64;
                                  var FCrc : cardinal) : integer;
var
  sComp        : TZDecompressionStream;
  NRead,NWrite : cardinal;
begin
  Result:=errOK;       // o.k.
    // init checksum
  FCrc:=Crc32(0,nil,0);
  // setup for GZip compatible decompression
  sComp:=TZDecompressionStream.Create(fSource,sLength,defWindowBits);
  repeat
    try
      NRead:=sComp.Read(FBuffer[0],FBufSize);
    except
      Result:=errFileRead; Exit;
      end;
    try
      FCrc:=ZipCrc32(FCrc,FBuffer[0],NRead);
      if assigned(fDest) then begin
        NWrite:=fDest.Write(FBuffer[0],NRead);
        inc(Prog,NWrite);
        end
      else begin
        inc(Prog,NRead);
        end;
      if Total>0 then pbUnzip.Position:=round(100/Total*Prog);
      Application.ProcessMessages;
    except
      Result:=errFileWrite; Exit;
      end;
    until (NRead<FBufSize) or Cancelled;
  sComp.Free;
  end;

function THauptForm.ReadZipFileHeader (sZip : TFileStream;
                                 var ZipInfo : TZipFileInfo) : boolean;
var
  LFHeader  : TPKLocalHeader;
  el,k      : cardinal;
  pkelf     : TExtraFieldInfo;
  Buffer    : array of byte;
  ZfName    : RawByteString;
begin
  sZip.Read(LFHeader,sizeof(TPKLocalHeader));         // local header
  with LFHeader do if Signatur=PKLocalSignatur then begin
    ZipInfo.TimeStamp:=FTimeStamp;
    ZipInfo.Crc:=Crc;
    ZipInfo.CSize:=CSize;
    ZipInfo.USize:=USize;
    ZipInfo.Flag:=Flag;
    ZipInfo.Method:=Method;
    ZipInfo.KeyMode:=1;      // default
    SetLength(ZfName,FNLength);
    sZip.Read(ZfName[1],FNLength);
    if Flag and zipUtf8Flag<>0 then ZipInfo.Filename:=UnixPathToDosPath(UTF8Decode(ZfName))
    else ZipInfo.Filename:=UnixPathToDosPath(StrToAnsi(ZfName));
    if ExtraLength>0 then begin      // read extra fields
      el:=ExtraLength;
      while el>0 do begin  // read extra fields
        sZip.Read(pkelf,SizeOf(TExtraFieldInfo));
        with pkelf do begin
          SetLength(Buffer,ESize);
          sZip.Read(Buffer[0],ESize);
          k:=0;
          if Tag=ExtraZip64 then begin
            if USize=MaxCardinal then begin
              Move(Buffer[k],ZipInfo.USize,8);
              inc(k,8);
              end;
            if (CSize=MaxCardinal) and (ESize>k) then begin
              Move(Buffer[k],ZipInfo.CSize,8);
              end
            end
          else if Tag=ExtraCrypt then begin
            ZipInfo.KeyMode:=Buffer[4];         // Mode
            end;
          el:=el-SizeOf(TExtraFieldInfo)-ESize;
          end;
        end;
      Buffer:=nil;
      end;
    Result:=true;
    end
  else Result:=false;
  end;

procedure THauptForm.ReadZip (FName : string);
var
  sZip      : TFileStream;
  rZip      : TZipFileInfo;
  fp        : int64;
  i,km      : integer;
  FCrc,n    : cardinal;
  ok        : boolean;
  mr        : TModalResult;
begin
  bnCancel.Enabled:=true;
  bnUpdate.Enabled:=false;
  sbLoadZip.Enabled:=false;
  bnSelectAll.Enabled:=false;
  bnUnzip.Enabled:=false;
  Cancelled:=false;
  Password:=''; km:=0;
  try
    sZip:=TFileStream.Create(FilenameToXL(FName),fmOpenRead+fmShareDenyNone);
  except
    ErrorDialog (_('Error opening Zip file:')+sLineBreak+FName);  //'Fehler beim ffnen der Zip-Datei:'
    bnCancel.Enabled:=false;
    bnUpdate.Enabled:=true;
    sbLoadZip.Enabled:=true;
    bnSelectAll.Enabled:=true;
    bnUnzip.Enabled:=true;
    exit;
    end;
  lbUnzip.Caption:=_('Scanning Zip file'); //'Durchsuche Zip-Datei'
  pbUnzip.Position:=0; fc:=0;
  Application.ProcessMessages;
  with lvFiles do begin
    for i:=0 to Items.Count-1 do with Items[i] do if Data<>nil then begin
      TPkFileRef(Data).Free;
      Data:=nil;
      end;
    Clear;
    end;
  repeat
    fp:=sZip.Position;
    ok:=ReadZipFileHeader (sZip,rZip);
    if ok and (rZip.CSize>0) then begin
      Total:=rZip.USize; Prog:=0;
      if cbCheck.Checked then begin
        lbUnzip.Caption:=_('Checking ')+StripPath(rZip.Filename,70); //'Prfe '
        if (rZip.Flag and zipEncryptFlag)<>0 then begin
          if not assigned(FEncryption) then begin
            mr:=PasswordDialog.Execute(CenterPos,Caption,Format(
              _('Checking of file:'+sLineBreak+'%s'+sLineBreak+
                'requires a password for decryption!'), //'Das berprfen der Datei:'+sLineBreak+'%s'+sLineBreak+'erfordert ein Kennwort zur Entschlsselung!',
              [StripPath(rZip.Filename,50)]),true,true,Password,km);
            if mr=mrOK then FEncryption:=TEncryption.Create(Password,km,defBufferSize)
            else if mr=mrCancel then Cancelled:=true;;
            end;
          if assigned(FEncryption) then ok:=FEncryption.DecryptStream(sZip,nil,rZip.CSize,false)
          else begin
            sZip.Seek (rZip.CSize,soFromCurrent);
            ok:=false;
            end;
          end
        else begin
          n:=GzUncompress (sZip,nil,rZip.CSize,FCrc);
          ok:=(n=errOk) and (FCrc=rZip.Crc);
          end;
        end
      else begin
        lbUnzip.Caption:=Format(_('Scanning Zip file (%u)'),[fc]); //'Durchsuche Zip-Datei ('
        sZip.Seek (rZip.CSize,soFromCurrent);
        ok:=true;
        end;
      with lvFiles do begin
        with Items.Add do begin
          Caption:=rZip.Filename;
          if ok then begin
            SubItems.Add(IntToStr(rZip.USize));
            SubItems.Add(FormatDateTime('dd.mm.yyyy hh:nn',FileDateToDateTime(rZip.TimeStamp)));
            Data:=TPkFileRef.Create(fp);
            end
          else begin
            SubItems.Add(_('Error')); //'Fehler'
            Data:=nil;
            end;
          MakeVisible(false);
          end;
        inc(fc);
        ok:=true;
        end;
      end;
    Application.ProcessMessages;
    until not ok or Cancelled;
  lbUnzip.Caption:=IntToStr(fc)+_(' file(s) found'); //' Datei(en) gefunden!'
  Application.ProcessMessages;
  sZip.Free;
  if assigned(FEncryption) then FreeAndNil(FEncryption);
  bnCancel.Enabled:=false;
  bnUpdate.Enabled:=true;
  sbLoadZip.Enabled:=true;
  bnSelectAll.Enabled:=true;
  bnUnzip.Enabled:=true;
  end;

procedure THauptForm.sbLoadZipClick(Sender: TObject);
begin
  with OpenDialog do begin
    if length(edZip.Text)>0 then InitialDir:=ExtractFilePath(edZip.Text)
    else InitialDir:=UserPath;
    Filename:='';
    if Execute then begin
      edZip.Text:=Filename;
      ReadZip(Filename);
      end;
    end;
  end;

procedure THauptForm.bnSelectAllClick(Sender: TObject);
begin
  lvFiles.SelectAll;
  end;

procedure THauptForm.bnUpdateClick(Sender: TObject);
begin
  ReadZip(edZip.Text);
  end;

procedure THauptForm.bnUnzipClick(Sender: TObject);
var
  i,n,km : integer;
  st,s  : string;
  sZip  : TFileStream;
  rZip  : TZipFileInfo;
  sTmp  : TStream;
  sDest : TFileStream;
  FCrc  : cardinal;
  ok    : boolean;
  mr    : TModalResult;
  tf    : string;
begin
  bnCancel.Enabled:=true;
  bnUpdate.Enabled:=false;
  sbLoadZip.Enabled:=false;
  bnSelectAll.Enabled:=false;
  bnUnzip.Enabled:=false;
  Cancelled:=false;
  Password:=''; tf:=''; km:=0;
  try
    sZip:=TFileStream.Create(FilenameToXL(edZip.Text),fmOpenRead+fmShareDenyNone);
  except
    ErrorDialog(_('Error opening Zip file:')+sLineBreak+edZip.Text); //'Fehler beim ffnen der Zip-Datei:'
    bnCancel.Enabled:=false;
    bnUpdate.Enabled:=true;
    sbLoadZip.Enabled:=true;
    bnSelectAll.Enabled:=true;
    bnUnzip.Enabled:=true;
    exit;
    end;
  pbUnzip.Position:=0;
  Application.ProcessMessages;
  fc:=0;
  if ShellDirDialog.Execute (_('Select destination directory'),true,true,false,'',Base) then begin  //'Ziel-Verzeichnis auswhlen'
    ForceDirectories(Base);
    with lvFiles do for i:=0 to Items.Count-1 do with Items[i] do
         if Selected and (Data<>nil) then begin
      sZip.Seek ((TPkFileRef(Data).Offset),soFromBeginning);
      if ReadZipFileHeader (sZip,rZip) then begin
        s:=AnsiReplaceStr(rZip.Filename,'/','\');
        st:=SetDirName(Base)+s;
        while length(st)>Max_Path do begin
          n:=AnsiPos('\',s);
          if n>0 then system.Delete(s,1,n);
          st:=SetDirName(Base)+s;
          end;
        ForceDirectories(ExtractFilePath(st));
        sDest:=TFileStream.Create(FilenameToXL(st),fmCreate);
        lbUnzip.Caption:=_('Unpacking: ')+StripPath(st,80); //'Entpacke: '
        Total:=rZip.USize; Prog:=0;
        if (rZip.Flag and zipEncryptFlag)<>0 then begin
          if not assigned(FEncryption) then begin
            mr:=PasswordDialog.Execute(CenterPos,Caption,Format(
              _('Unpacking of file:'+sLineBreak+'%s'+sLineBreak+
                'requires a password for decryption!'), //'Das Entpackender Datei:'+sLineBreak+'%s'+sLineBreak+'erfordert ein Kennwort zur Entschlsselung!',
              [StripPath(rZip.Filename,50)]),true,true,Password,km);
            if mr=mrOK then FEncryption:=TEncryption.Create(Password,km,defBufferSize)
            else if mr=mrCancel then Cancelled:=true;;
            end;
          if assigned(FEncryption) then begin
            if rZip.CSize>MaxMemSize then begin
              tf:=SetDirName(Base)+TempName;
              sTmp:=TFileStream.Create(tf,fmCreate); // write to disk
              end
            else sTmp:=TMemoryStream.Create;
            ok:=FEncryption.DecryptStream(sZip,sTmp,rZip.CSize,false);
            if ok then begin
              sTmp.Seek(0,soFromBeginning);
              ok:=GzUncompress(sTmp,sDest,sTmp.size,FCrc)=errOK;
              end;
            sTmp.Free;
            end
          else begin
            sZip.Seek (rZip.CSize,soFromCurrent);
            ok:=false;
            end;
          end
        else begin
          n:=GzUncompress (sZip,sDest,rZip.CSize,FCrc);
          ok:=(n=errOK) and (FCrc=rZip.Crc);
          end;
        sDest.Free;
        if ok then inc(fc)
        else ErrorDialog(_('Error on unpacking file:')+sLineBreak+rZip.Filename); //'Fehler beim Entpacken von Datei:'
        n:=FileSetDate(st,rZip.TimeStamp);
        if n=0 then FileSetAttr(st,faArchive);
//        if n<>0 then ShowError ('Fehler beim Setzen der Attribute:'sLineBreak+rZip.Filename);
        end
      else ErrorDialog(Format(_('Header of file:'+sLineBreak+'%s'+sLineBreak+'not found!'),[rZip.Filename])); //'Header fr Datei :+%s+nicht gefunden!'
      if Cancelled then Break;
      end;
    end;
  lbUnzip.Caption:=IntToStr(fc)+_(' file(s) unpacked'); //' Datei(en) entpackt!'
  Application.ProcessMessages;
  sZip.Free;
  if assigned(FEncryption) then FreeAndNil(FEncryption);
  if (length(tf)>0) and FileExists(tf) then DeleteFile(tf);
  bnCancel.Enabled:=false;
  bnUpdate.Enabled:=true;
  sbLoadZip.Enabled:=true;
  bnSelectAll.Enabled:=true;
  bnUnzip.Enabled:=true;
  end;

procedure THauptForm.bnEndeClick(Sender: TObject);
begin
  Cancelled:=true;
  Close;
  end;

procedure THauptForm.bnCancelClick(Sender: TObject);
begin
  Cancelled:=true;
  end;

procedure THauptForm.bnInfoClick(Sender: TObject);
begin
  InfoDialog (ProgName,Caption+' - '+ProgVersDate+sLineBreak+
           VersInfo.CopyRight+sLineBreak+
           _('The program unpacks Zip archives without using'+sLineBreak+
           'the central directory at end of file')); //'Das Programm entpackt Zip-Archive ohne Benutzung'+sLineBreak+'des Verzeichnisses am Ende der Datei',

  end;

end.
